@@ -281,7 +281,7 @@ class AtlasClient:
             existing_project = search_results[0]
             return existing_project
         else:
-            raise Exception("Could not find project `{project_name} in organization `{organization_name}``")
+            raise Exception(f"Could not find project `{project_name} in organization `{organization_name}``")
 
     def _get_project_by_id(self, project_id: str):
         '''
@@ -530,44 +530,73 @@ class AtlasClient:
                 headers=self.header,
                 json={'project_id': project_id, 'embeddings': base64.b64encode(bytesio.getvalue()).decode('utf-8'), 'data': data_shard},
             )
-            del embedding_shard
             return response
 
-        failed = []
 
         # if this method is being called internally, we pass a global progress bar
         close_pbar = False
         if pbar is None:
-            logger.info("Uploading embeddings to Nomic.")
+            logger.info("Uploading embeddings to Atlas.")
             close_pbar = True
             pbar = tqdm(total=int(embeddings.shape[0]) // shard_size)
-
+        failed = 0
+        succeeded = 0
+        errors_504 = 0
         with concurrent.futures.ThreadPoolExecutor(max_workers=num_workers) as executor:
             futures = {executor.submit(send_request, i): i for i in range(0, len(data), shard_size)}
-            for future in concurrent.futures.as_completed(futures):
-                response = future.result()
-                pbar.update(1)
-                if response.status_code != 200:
-                    try:
-                        logger.error(f"Shard upload failed: {response.json()}")
-                        if 'more datums exceeds your organization limit' in response.json():
-                            return False
-                        if 'Project transaction lock is held':
-                            raise Exception("Project is currently indexing and cannot ingest new datums. Try again later.")
-                    except (requests.JSONDecodeError, json.decoder.JSONDecodeError):
-                        if response.status_code == 413:
-                            logger.error("Shard upload failed: you are sending meta-data data is to large.")
-                        else:
-                            logger.error(f"Shard upload failed: {response}")
-                        continue
 
-                    failed.append(futures[future])
+            while futures:
+                # check for status of the futures which are currently working
+                done, not_done = concurrent.futures.wait(
+                    futures,
+                    return_when=concurrent.futures.FIRST_COMPLETED)
+                # process any completed futures
+                for future in done:
+                    response = future.result()
+                    if response.status_code != 200:
+                        try:
+                            logger.error(f"Shard upload failed: {response.json()}")
+                            if 'more datums exceeds your organization limit' in response.json():
+                                return False
+                            if 'Project transaction lock is held':
+                                raise Exception("Project is currently indexing and cannot ingest new datums. Try again later.")
+                        except (requests.JSONDecodeError, json.decoder.JSONDecodeError):
+                            if response.status_code == 413:
+                                # Possibly split in two and retry?
+                                logger.error("Shard upload failed: you are sending meta-data that is too large.")
+                                pbar.update(1)
+                                response.close()
+                                failed += shard_size
+                            elif response.status_code == 504:
+                                errors_504 += shard_size
+                                start_point = futures[future]
+                                logger.debug(f"Connection failed for records {start_point}-{start_point + shard_size}, retrying.")
+                                failure_fraction = errors_504 / (failed + succeeded + errors_504)
+                                if failure_fraction > 0.25 and errors_504 > shard_size * 3:
+                                    raise RuntimeError("Atlas is under high load and cannot ingest datums at this time. Please try again later.")
+                                new_submission = executor.submit(send_request, start_point)
+                                futures[new_submission] = start_point
+                                response.close()
+                            else:
+                                logger.error(f"Shard upload failed: {response}")
+                                failed += shard_size
+                                pbar.update(1)
+                                response.close()
+                    else:
+                        # A successful upload.
+                        succeeded += shard_size
+                        pbar.update(1)
+                        response.close()
+
+                    # remove the now completed future
+                    del futures[future]
+
         # close the progress bar if this method was called with no external progresbar
         if close_pbar:
             pbar.close()
 
         if failed:
-            logger.warning(f"Failed to upload {len(failed)*shard_size} datums")
+            logger.warning(f"Failed to upload {failed} datums")
         if close_pbar:
             if failed:
                 logger.warning("Embedding upload partially succeeded.")
@@ -585,7 +614,8 @@ class AtlasClient:
                      projection_n_neighbors=DEFAULT_PROJECTION_N_NEIGHBORS,
                      projection_epochs=DEFAULT_PROJECTION_EPOCHS,
                      projection_spread=DEFAULT_PROJECTION_SPREAD,
-                     topic_label_field = None
+                     topic_label_field = None,
+                     reuse_lm_from = None
                      ) -> CreateIndexResponse:
         '''
         Creates an index in the specified project
@@ -600,6 +630,7 @@ class AtlasClient:
         * **shard_size** - Embeddings are uploaded in parallel by many threads. Adjust the number of embeddings to upload by each worker.
         * **num_workers** - The number of worker threads to upload embeddings with.
         * **topic_label_field** - A text field to estimate topic labels from.
+        * **reuse_lm_from** - An optional index id from the same project whose atoms and embeddings should be reused. Text projects only.
 
         **Returns:** A link to your map.
         '''
@@ -655,6 +686,7 @@ class AtlasClient:
                 'dataset_buffer_size': 1000,
                 'batch_size': 20,
                 'polymerize_by': 'charchunk',
+                'norm': 'both',
             }
             text_build_template = {
                 'project_id': project['id'],
@@ -663,6 +695,7 @@ class AtlasClient:
                 'atomizer_strategies': ['document', 'charchunk'],
                 'model': model,
                 'colorable_fields': colorable_fields,
+                'reuse_atoms_and_embeddings_from': reuse_lm_from,
                 'model_hyperparameters': json.dumps(hyperparameters),
                 'nearest_neighbor_index': 'HNSWIndex',
                 'nearest_neighbor_index_hyperparameters': json.dumps({'space': 'l2', 'ef_construction': 100, 'M': 16}),
@@ -792,27 +825,61 @@ class AtlasClient:
         # if this method is being called internally, we pass a global progress bar
         close_pbar = False
         if pbar is None:
-            logger.info("Uploading text to Nomic.")
+            logger.info("Uploading text to Atlas.")
             close_pbar = True
             pbar = tqdm(total=int(len(data)) // shard_size)
-
+        failed = 0
+        succeeded = 0
+        errors_504 = 0
         with concurrent.futures.ThreadPoolExecutor(max_workers=num_workers) as executor:
             futures = {executor.submit(send_request, i): i for i in range(0, len(data), shard_size)}
-            for future in concurrent.futures.as_completed(futures):
-                response = future.result()
-                pbar.update(1)
-                if response.status_code != 200:
-                    try:
-                        logger.error(f"Shard upload failed: {response.json()}")
-                        if 'more datums exceeds your organization limit' in response.json():
-                            return False
-                        if 'Project transaction lock is held':
-                            raise Exception("Project is currently indexing and cannot ingest new datums. Try again later.")
-                    except requests.exceptions.JSONDecodeError:
-                        logger.error(f"Shard upload failed: {response}")
-                        continue
 
-                    failed.append(futures[future])
+            while futures:
+                # check for status of the futures which are currently working
+                done, not_done = concurrent.futures.wait(
+                    futures,
+                    return_when=concurrent.futures.FIRST_COMPLETED)
+                # process any completed futures
+                for future in done:
+                    response = future.result()
+                    if response.status_code != 200:
+                        try:
+                            logger.error(f"Shard upload failed: {response.json()}")
+                            if 'more datums exceeds your organization limit' in response.json():
+                                return False
+                            if 'Project transaction lock is held':
+                                raise Exception("Project is currently indexing and cannot ingest new datums. Try again later.")
+                        except (requests.JSONDecodeError, json.decoder.JSONDecodeError):
+                            if response.status_code == 413:
+                                # Possibly split in two and retry?
+                                logger.error("Shard upload failed: you are sending meta-data that is too large.")
+                                pbar.update(1)
+                                response.close()
+                                failed += shard_size
+                            elif response.status_code == 504:
+                                errors_504 += shard_size
+                                start_point = futures[future]
+                                logger.debug(f"Connection failed for records {start_point}-{start_point + shard_size}, retrying.")
+                                failure_fraction = errors_504 / (failed + succeeded + errors_504)
+                                if failure_fraction > 0.25 and errors_504 > shard_size * 3:
+                                    raise RuntimeError("Atlas is under high load and cannot ingest datums at this time. Please try again later.")
+                                new_submission = executor.submit(send_request, start_point)
+                                futures[new_submission] = start_point
+                                response.close()
+                            else:
+                                logger.error(f"Shard upload failed: {response}")
+                                failed += shard_size
+                                pbar.update(1)
+                                response.close()
+                    else:
+                        # A successful upload.
+                        succeeded += shard_size
+                        pbar.update(1)
+                        response.close()
+
+                    # remove the now completed future
+                    del futures[future]
+
         # close the progress bar if this method was called with no external progresbar
         if close_pbar:
             pbar.close()
@@ -905,7 +972,7 @@ class AtlasClient:
         number_of_datums_before_upload = project['total_datums_in_project']
 
         # sends several requests to allow for threadpool refreshing. Threadpool hogs memory and new ones need to be created.
-        logger.info("Uploading embeddings to Nomic's neural database Atlas.")
+        logger.info("Uploading embeddings to Atlas.")
 
         embeddings = embeddings.astype(np.float16)
         with tqdm(total=len(data) // shard_size) as pbar:
@@ -1048,7 +1115,7 @@ class AtlasClient:
         number_of_datums_before_upload = project['total_datums_in_project']
 
 
-        logger.info("Uploading text to Nomic's neural database Atlas.")
+        logger.info("Uploading text to Atlas.")
 
         with tqdm(total=len(data) // shard_size) as pbar:
             for i in range(0, len(data), MAX_MEMORY_CHUNK):
